@@ -29,6 +29,7 @@
 #define PKCS11_ENGINE_NAME "pkcs11 engine"
 
 static int pkcs11_idx = -1;
+static int shutdown_mode = 0;
 
 /* The definitions for control commands specific to this engine */
 
@@ -82,6 +83,8 @@ static const ENGINE_CMD_DEFN engine_cmd_defns[] = {
 	{0, NULL, NULL, 0}
 };
 
+static int bind_helper_methods(ENGINE *e);
+
 static ENGINE_CTX *get_ctx(ENGINE *engine)
 {
 	ENGINE_CTX *ctx;
@@ -111,17 +114,6 @@ static int engine_destroy(ENGINE *engine)
 	if (!ctx)
 		return 0;
 
-	/* ENGINE_remove() invokes our engine_destroy() function with
-	 * CRYPTO_LOCK_ENGINE / global_engine_lock acquired.
-	 * Any attempt to re-acquire the lock either by directly
-	 * invoking OpenSSL functions, or indirectly via PKCS#11 modules
-	 * that use OpenSSL engines, causes a deadlock. */
-	/* Our workaround is to skip ctx_finish() entirely, as a memory
-	 * leak is better than a deadlock. */
-#if 0
-	rv &= ctx_finish(ctx);
-#endif
-
 	rv &= ctx_destroy(ctx);
 	ENGINE_set_ex_data(engine, pkcs11_idx, NULL);
 	ERR_unload_ENG_strings();
@@ -148,20 +140,14 @@ static int engine_finish(ENGINE *engine)
 	if (!ctx)
 		return 0;
 
-	/* ENGINE_cleanup() used by OpenSSL versions before 1.1.0 invokes
-	 * our engine_finish() function with CRYPTO_LOCK_ENGINE acquired.
-	 * Any attempt to re-acquire CRYPTO_LOCK_ENGINE either by directly
-	 * invoking OpenSSL functions, or indirectly via PKCS#11 modules
-	 * that use OpenSSL engines, causes a deadlock. */
-	/* Our workaround is to skip ctx_finish() for the affected OpenSSL
-	 * versions, as a memory leak is better than a deadlock. */
-	/* We cannot simply temporarily release CRYPTO_LOCK_ENGINE here, as
-	 * engine_finish() is also executed from ENGINE_finish() without
-	 * acquired CRYPTO_LOCK_ENGINE, and there is no way with to check
-	 * whether a lock is already acquired with OpenSSL < 1.1.0 API. */
-#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
-	rv &= ctx_finish(ctx);
-#endif
+	/* PKCS#11 modules that register their own atexit() callbacks may
+	 * already have been cleaned up by the time OpenSSL's atexit() callback
+	 * is executed. As a result, a crash occurs with certain versions of
+	 * OpenSSL and SoftHSM2. The workaround skips the execution of
+	 * ctx_finish() during OpenSSL's cleanup, converting the crash into
+	 * a harmless memory leak at exit. */
+	if (!shutdown_mode)
+		rv &= ctx_finish(ctx);
 
 	return rv;
 }
@@ -174,6 +160,7 @@ static EVP_PKEY *load_pubkey(ENGINE *engine, const char *s_key_id,
 	ctx = get_ctx(engine);
 	if (!ctx)
 		return 0;
+	bind_helper_methods(engine);
 	return ctx_load_pubkey(ctx, s_key_id, ui_method, callback_data);
 }
 
@@ -186,6 +173,33 @@ static EVP_PKEY *load_privkey(ENGINE *engine, const char *s_key_id,
 	ctx = get_ctx(engine);
 	if (!ctx)
 		return 0;
+	bind_helper_methods(engine);
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	/*
+	 * A workaround for an OpenSSL bug affecting the handling of foreign
+	 * EVP_PKEY objects: https://github.com/openssl/openssl/pull/23063
+	 * Affected OpenSSL versions:
+	 *  - 3.0.12 (0x300000c0L) - 3.0.13 (0x300000d0L)
+	 *  - 3.1.4 (0x30100040L) - 3.1.5 (0x30100050L)
+	 *  - 3.2.0 (0x30200000L) - 3.2.1 (0x30200010L)
+	 * This workaround may disrupt rare deployments
+	 * that use foreign keys from multiple engines.
+	 */
+	{
+		unsigned long ver = OpenSSL_version_num();
+
+		if ((ver >= 0x300000c0L && ver <= 0x300000d0L) ||
+				(ver >= 0x30100040L && ver <= 0x30100050L) ||
+				(ver >= 0x30200000L && ver <= 0x30200010L)) {
+			if (ENGINE_set_default_string(engine, "PKEY_CRYPTO")) {
+				fprintf(stderr, "Workaround for %s enabled\n",
+					OpenSSL_version(OPENSSL_VERSION));
+			} else {
+				fprintf(stderr, "Failed to set PKEY_CRYPTO default engine\n");
+			}
+		}
+	}
+#endif
 	pkey = ctx_load_privkey(ctx, s_key_id, ui_method, callback_data);
 #ifdef EVP_F_EVP_PKEY_SET1_ENGINE
 	/* EVP_PKEY_set1_engine() is required for OpenSSL 1.1.x,
@@ -205,6 +219,9 @@ static int engine_ctrl(ENGINE *engine, int cmd, long i, void *p, void (*f) ())
 	ctx = get_ctx(engine);
 	if (!ctx)
 		return 0;
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
+	bind_helper_methods(engine);
+#endif
 	return ctx_engine_ctrl(ctx, cmd, i, p, f);
 }
 
@@ -219,10 +236,28 @@ static int bind_helper(ENGINE *e)
 			!ENGINE_set_ctrl_function(e, engine_ctrl) ||
 			!ENGINE_set_cmd_defns(e, engine_cmd_defns) ||
 			!ENGINE_set_name(e, PKCS11_ENGINE_NAME) ||
+			!ENGINE_set_load_pubkey_function(e, load_pubkey) ||
+			!ENGINE_set_load_privkey_function(e, load_privkey)) {
+		return 0;
+	} else {
+		ERR_load_ENG_strings();
+		return 1;
+	}
+}
+
+/*
+ * With OpenSSL 3.x, engines might be used because defined in openssl.cnf
+ * which will cause problems
+ * only add engine routines after a call to load keys
+ */
+
+static int bind_helper_methods(ENGINE *e)
+{
+	if (
 #ifndef OPENSSL_NO_RSA
 			!ENGINE_set_RSA(e, PKCS11_get_rsa_method()) ||
 #endif
-#if OPENSSL_VERSION_NUMBER  >= 0x10100002L
+#if OPENSSL_VERSION_NUMBER >= 0x10100002L
 #ifndef OPENSSL_NO_EC
 			/* PKCS11_get_ec_key_method combines ECDSA and ECDH */
 			!ENGINE_set_EC(e, PKCS11_get_ec_key_method()) ||
@@ -235,16 +270,18 @@ static int bind_helper(ENGINE *e)
 			!ENGINE_set_ECDH(e, PKCS11_get_ecdh_method()) ||
 #endif
 #endif /* OPENSSL_VERSION_NUMBER */
-			!ENGINE_set_pkey_meths(e, PKCS11_pkey_meths) ||
-			!ENGINE_set_load_pubkey_function(e, load_pubkey) ||
-			!ENGINE_set_load_privkey_function(e, load_privkey)) {
+			!ENGINE_set_pkey_meths(e, PKCS11_pkey_meths)) {
 		return 0;
 	} else {
-		ERR_load_ENG_strings();
 		return 1;
 	}
 }
 
+static void exit_callback(void)
+{
+	shutdown_mode = 1;
+}
+
 static int bind_fn(ENGINE *e, const char *id)
 {
 	if (id && (strcmp(id, PKCS11_ENGINE_ID) != 0)) {
@@ -255,6 +292,7 @@ static int bind_fn(ENGINE *e, const char *id)
 		fprintf(stderr, "bind failed\n");
 		return 0;
 	}
+	atexit(exit_callback);
 	return 1;
 }
 
