@@ -24,6 +24,8 @@
 #include "config.h"
 #endif
 
+/* this code extensively uses deprecated features, so warnings are useless */
+#define OPENSSL_SUPPRESS_DEPRECATED
 #include "libp11.h"
 
 #define CRYPTOKI_EXPORTS
@@ -46,6 +48,7 @@ struct pkcs11_ctx_private {
 	CK_FUNCTION_LIST_PTR method;
 	void *handle;
 	char *init_args;
+	CK_VERSION cryptoki_version;
 	UI_METHOD *ui_method; /* UI_METHOD for CKU_CONTEXT_SPECIFIC PINs */
 	void *ui_user_data;
 	unsigned int forkid;
@@ -93,6 +96,8 @@ struct pkcs11_object_private {
 	EVP_PKEY *evp_key;
 	X509 *x509;
 	unsigned int forkid;
+	int refcnt;
+	pthread_mutex_t lock;
 };
 #define PRIVKEY(_key)		((PKCS11_OBJECT_private *) (_key)->_private)
 #define PRIVCERT(_cert)		((PKCS11_OBJECT_private *) (_cert)->_private)
@@ -207,7 +212,7 @@ extern int pkcs11_get_session(PKCS11_SLOT_private *, int rw, CK_SESSION_HANDLE *
 extern void pkcs11_put_session(PKCS11_SLOT_private *, CK_SESSION_HANDLE session);
 
 /* Get a list of all slots */
-extern int pkcs11_enumerate_slots(PKCS11_CTX_private * ctx,
+extern int pkcs11_enumerate_slots(PKCS11_CTX_private *ctx,
 			PKCS11_SLOT **slotsp, unsigned int *nslotsp);
 
 /* Get the slot_id from a slot as it is stored in private */
@@ -216,8 +221,9 @@ extern unsigned long pkcs11_get_slotid_from_slot(PKCS11_SLOT_private *);
 /* Increment slot reference count */
 extern PKCS11_SLOT_private *pkcs11_slot_ref(PKCS11_SLOT_private *slot);
 
-/* Decrement slot reference count, free if it becomes zero */
-extern void pkcs11_slot_unref(PKCS11_SLOT_private *slot);
+/* Decrement slot reference count, free if it becomes zero.
+ * Returns 1 if it was freed. */
+extern int pkcs11_slot_unref(PKCS11_SLOT_private *slot);
 
 /* Free the list of slots allocated by PKCS11_enumerate_slots() */
 extern void pkcs11_release_all_slots(PKCS11_SLOT *slots, unsigned int nslots);
@@ -237,9 +243,9 @@ extern int pkcs11_logout(PKCS11_SLOT_private *);
 /* Authenticate a private the key operation if needed */
 int pkcs11_authenticate(PKCS11_OBJECT_private *key, CK_SESSION_HANDLE session);
 
-/* Get a list of keys associated with this token */
+/* Get a list of keys matching with template associated with this token */
 extern int pkcs11_enumerate_keys(PKCS11_SLOT_private *, unsigned int type,
-	PKCS11_KEY **keys, unsigned int *nkeys);
+	const PKCS11_KEY *key_template, PKCS11_KEY **keys, unsigned int *nkeys);
 
 /* Create an object from a handle */
 extern PKCS11_OBJECT_private *pkcs11_object_from_handle(PKCS11_SLOT_private *slot,
@@ -253,6 +259,9 @@ extern PKCS11_OBJECT_private *pkcs11_object_from_template(PKCS11_SLOT_private *s
 extern PKCS11_OBJECT_private *pkcs11_object_from_object(PKCS11_OBJECT_private *obj,
 	CK_SESSION_HANDLE session, CK_OBJECT_CLASS object_class);
 
+/* Reference the private object */
+extern PKCS11_OBJECT_private *pkcs11_object_ref(PKCS11_OBJECT_private *obj);
+
 /* Free an object */
 extern void pkcs11_object_free(PKCS11_OBJECT_private *obj);
 
@@ -268,9 +277,9 @@ extern PKCS11_CERT *pkcs11_find_certificate(PKCS11_OBJECT_private *key);
 /* Find the corresponding key (if any) */
 extern PKCS11_KEY *pkcs11_find_key(PKCS11_OBJECT_private *cert);
 
-/* Get a list of all certificates associated with this token */
+/* Get a list of all certificates matching with template associated with this token */
 extern int pkcs11_enumerate_certs(PKCS11_SLOT_private *,
-	PKCS11_CERT **certs, unsigned int *ncerts);
+	const PKCS11_CERT *cert_template, PKCS11_CERT **certs, unsigned int *ncerts);
 
 /* Remove an object from the token */
 extern int pkcs11_remove_object(PKCS11_OBJECT_private *object);
@@ -299,7 +308,7 @@ extern int pkcs11_store_public_key(PKCS11_SLOT_private *,
 	EVP_PKEY *pk, char *label, unsigned char *id, size_t id_len);
 
 /* Store certificate on a token */
-extern int pkcs11_store_certificate(PKCS11_SLOT_private *, X509 * x509,
+extern int pkcs11_store_certificate(PKCS11_SLOT_private *, X509 *x509,
 		char *label, unsigned char *id, size_t id_len,
 		PKCS11_CERT **ret_cert);
 
@@ -312,7 +321,7 @@ extern int pkcs11_generate_random(PKCS11_SLOT_private *, unsigned char *r, unsig
 /* Generate and store a private key on the token */
 extern int pkcs11_generate_key(PKCS11_SLOT_private *tpriv,
 	int algorithm, unsigned int bits,
-	char *label, unsigned char* id, size_t id_len);
+	char *label, unsigned char *id, size_t id_len);
 
 /* Get the RSA key modulus size (in bytes) */
 extern int pkcs11_get_key_size(PKCS11_OBJECT_private *);
@@ -346,9 +355,27 @@ extern int pkcs11_private_decrypt(
 /* Retrieve PKCS11_KEY from an RSA key */
 extern PKCS11_OBJECT_private *pkcs11_get_ex_data_rsa(const RSA *rsa);
 
+/* Set PKCS11_KEY for an RSA key */
+void pkcs11_set_ex_data_rsa(RSA *rsa, PKCS11_OBJECT_private *key);
+
 /* Retrieve PKCS11_KEY from an EC_KEY */
 extern PKCS11_OBJECT_private *pkcs11_get_ex_data_ec(const EC_KEY *ec);
 
+/* Set PKCS11_KEY for an EC_KEY */
+extern void pkcs11_set_ex_data_ec(EC_KEY *ec, PKCS11_OBJECT_private *key);
+
+/* Free the global RSA_METHOD */
+extern void pkcs11_rsa_method_free(void);
+
+/* Free the global EC_KEY_METHOD */
+extern void pkcs11_ec_key_method_free(void);
+
+/* Free the global ECDSA_METHOD */
+extern void pkcs11_ecdsa_method_free(void);
+
+/* Free the global ECDH_METHOD */
+extern void pkcs11_ecdh_method_free(void);
+
 #endif
 
 /* vim: set noexpandtab: */
