@@ -24,31 +24,28 @@
 #ifndef _LIB11_H
 #define _LIB11_H
 
+#include "p11_err.h"
 #include <openssl/bio.h>
 #include <openssl/err.h>
 #include <openssl/bn.h>
 #include <openssl/rsa.h>
 #include <openssl/x509.h>
+#include <openssl/evp.h>
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-/* Get some structures for local code to handle PKCS#11 data readily */
-#define ERR_LIB_PKCS11	ERR_LIB_USER
-
-#define PKCS11err(f,r) \
-	ERR_PUT_error(ERR_LIB_PKCS11,(f),(r),__FILE__,__LINE__)
+int ERR_load_CKR_strings(void);
+void ERR_unload_CKR_strings(void);
+void ERR_CKR_error(int function, int reason, char *file, int line);
+# define CKRerr(f,r) ERR_CKR_error((f),(r),__FILE__,__LINE__)
+int ERR_get_CKR_code(void);
 
 /*
- * The purpose of this library is to provide a simple PKCS11
- * interface to OpenSSL application that wish to use a previously
- * initialized card (as opposed to initializing it, etc).
- *
- * I am therefore making some simplifying assumptions:
- *
- *  -	no support for any operations that alter the card,
- *  	i.e. readonly-login
+ * The purpose of this library is to provide a simple PKCS#11
+ * interface to OpenSSL application.  It was never a goal
+ * of this project to expose the entire PKCS#11 functionality.
  */
 
 /** PKCS11 key object (public or private) */
@@ -58,7 +55,6 @@ typedef struct PKCS11_key_st {
 	size_t id_len;
 	unsigned char isPrivate;	/**< private key present? */
 	unsigned char needLogin;	/**< login to read private key? */
-	EVP_PKEY *evp_key;		/**< initially NULL, need to call PKCS11_load_key */
 	void *_private;
 } PKCS11_KEY;
 
@@ -91,7 +87,7 @@ typedef struct PKCS11_token_st {
 	unsigned char soPinFinalTry;
 	unsigned char soPinLocked;
 	unsigned char soPinToBeChanged;
-	void *_private;
+	struct PKCS11_slot_st *slot;
 } PKCS11_TOKEN;
 
 /** PKCS11 slot: card reader */
@@ -123,7 +119,7 @@ extern PKCS11_CTX *PKCS11_CTX_new(void);
  *
  * @return none
  */
-extern void PKCS11_CTX_init_args(PKCS11_CTX * ctx, const char * init_args);
+extern void PKCS11_CTX_init_args(PKCS11_CTX *ctx, const char *init_args);
 
 /**
  * Load a PKCS#11 module
@@ -133,30 +129,21 @@ extern void PKCS11_CTX_init_args(PKCS11_CTX * ctx, const char * init_args);
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_CTX_load(PKCS11_CTX * ctx, const char * ident);
-
-/**
- * Reinitialize a PKCS#11 module (after a fork)
- *
- * @param ctx context allocated by PKCS11_CTX_new()
- * @retval 0 success
- * @retval -1 error
- */
-extern int PKCS11_CTX_reload(PKCS11_CTX * ctx);
+extern int PKCS11_CTX_load(PKCS11_CTX *ctx, const char *ident);
 
 /**
  * Unload a PKCS#11 module
  *
  * @param ctx context allocated by PKCS11_CTX_new()
  */
-extern void PKCS11_CTX_unload(PKCS11_CTX * ctx);
+extern void PKCS11_CTX_unload(PKCS11_CTX *ctx);
 
 /**
  * Free a libp11 context
  *
  * @param ctx context allocated by PKCS11_CTX_new()
  */
-extern void PKCS11_CTX_free(PKCS11_CTX * ctx);
+extern void PKCS11_CTX_free(PKCS11_CTX *ctx);
 
 /** Open a session in RO or RW mode
  *
@@ -165,7 +152,7 @@ extern void PKCS11_CTX_free(PKCS11_CTX * ctx);
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_open_session(PKCS11_SLOT * slot, int rw);
+extern int PKCS11_open_session(PKCS11_SLOT *slot, int rw);
 
 /**
  * Get a list of all slots
@@ -176,7 +163,24 @@ extern int PKCS11_open_session(PKCS11_SLOT * slot, int rw);
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_enumerate_slots(PKCS11_CTX * ctx,
+extern int PKCS11_enumerate_slots(PKCS11_CTX *ctx,
+			PKCS11_SLOT **slotsp, unsigned int *nslotsp);
+
+/**
+ * Get or update a list of all slots
+ *
+ * The difference to PKCS11_enumerate_slots() is that this will
+ * expect as input previous slot list (or zero initialized count
+ * and null pointer) for the list. This function always reuses the
+ * slots found from the previous list to avoid unexpected slot
+ * and key object destructon.
+ * @param ctx context allocated by PKCS11_CTX_new()
+ * @param slotsp pointer on a list of slots
+ * @param nslotsp pointer to size of the allocated list
+ * @retval 0 success
+ * @retval -1 error
+ */
+extern int PKCS11_update_slots(PKCS11_CTX *ctx,
 			PKCS11_SLOT **slotsp, unsigned int *nslotsp);
 
 /**
@@ -194,7 +198,7 @@ extern unsigned long PKCS11_get_slotid_from_slot(PKCS11_SLOT *slotp);
  * @param slots list of slots allocated by PKCS11_enumerate_slots()
  * @param nslots size of the list
  */
-extern void PKCS11_release_all_slots(PKCS11_CTX * ctx,
+extern void PKCS11_release_all_slots(PKCS11_CTX *ctx,
 			PKCS11_SLOT *slots, unsigned int nslots);
 
 /**
@@ -206,9 +210,23 @@ extern void PKCS11_release_all_slots(PKCS11_CTX * ctx,
  * @retval !=NULL pointer on a slot structure
  * @retval NULL error
  */
-PKCS11_SLOT *PKCS11_find_token(PKCS11_CTX * ctx,
+PKCS11_SLOT *PKCS11_find_token(PKCS11_CTX *ctx,
 			PKCS11_SLOT *slots, unsigned int nslots);
 
+/**
+ * Find the next slot with a token
+ *
+ * @param ctx context allocated by PKCS11_CTX_new()
+ * @param slots list of slots allocated by PKCS11_enumerate_slots()
+ * @param nslots size of the list
+ * @param slot current slot
+ * @retval !=NULL pointer on a slot structure
+ * @retval NULL error
+ */
+PKCS11_SLOT *PKCS11_find_next_token(PKCS11_CTX *ctx,
+			PKCS11_SLOT *slots, unsigned int nslots,
+		   	PKCS11_SLOT *slot);
+
 /**
  * Check if user is already authenticated to a card
  *
@@ -218,7 +236,7 @@ PKCS11_SLOT *PKCS11_find_token(PKCS11_CTX * ctx,
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_is_logged_in(PKCS11_SLOT * slot, int so, int * res);
+extern int PKCS11_is_logged_in(PKCS11_SLOT *slot, int so, int *res);
 
 /**
  * Authenticate to the card
@@ -229,7 +247,7 @@ extern int PKCS11_is_logged_in(PKCS11_SLOT * slot, int so, int * res);
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_login(PKCS11_SLOT * slot, int so, const char *pin);
+extern int PKCS11_login(PKCS11_SLOT *slot, int so, const char *pin);
 
 /**
  * De-authenticate from the card
@@ -238,16 +256,27 @@ extern int PKCS11_login(PKCS11_SLOT * slot, int so, const char *pin);
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_logout(PKCS11_SLOT * slot);
+extern int PKCS11_logout(PKCS11_SLOT *slot);
 
 /* Get a list of private keys associated with this token */
 extern int PKCS11_enumerate_keys(PKCS11_TOKEN *,
 	PKCS11_KEY **, unsigned int *);
 
+/* Get a list of private keys associated with this token and matching the key template */
+extern int PKCS11_enumerate_keys_ext(PKCS11_TOKEN *,
+	const PKCS11_KEY *, PKCS11_KEY **, unsigned int *);
+
+/* Remove the key from this token */
+extern int PKCS11_remove_key(PKCS11_KEY *);
+
 /* Get a list of public keys associated with this token */
 extern int PKCS11_enumerate_public_keys(PKCS11_TOKEN *,
 	PKCS11_KEY **, unsigned int *);
 
+/* Get a list of public keys associated with this token and matching the key template */
+extern int PKCS11_enumerate_public_keys_ext(PKCS11_TOKEN *,
+	const PKCS11_KEY *, PKCS11_KEY **, unsigned int *);
+
 /* Get the key type (as EVP_PKEY_XXX) */
 extern int PKCS11_get_key_type(PKCS11_KEY *);
 
@@ -255,7 +284,7 @@ extern int PKCS11_get_key_type(PKCS11_KEY *);
  * Returns a EVP_PKEY object for the private key
  *
  * @param   key  PKCS11_KEY object
- * @retval !=NULL reference to EVP_PKEY object.
+ * @retval !=NULL reference to the EVP_PKEY object
  * @retval NULL error
  */
 extern EVP_PKEY *PKCS11_get_private_key(PKCS11_KEY *key);
@@ -264,7 +293,7 @@ extern EVP_PKEY *PKCS11_get_private_key(PKCS11_KEY *key);
  * Returns a EVP_PKEY object with the public key
  *
  * @param  key  PKCS11_KEY object
- * @retval !=NULL reference to EVP_PKEY object.
+ * @retval !=NULL reference to the EVP_PKEY object
  * @retval NULL error
  */
 extern EVP_PKEY *PKCS11_get_public_key(PKCS11_KEY *key);
@@ -275,12 +304,16 @@ extern PKCS11_CERT *PKCS11_find_certificate(PKCS11_KEY *);
 /* Find the corresponding key (if any) */
 extern PKCS11_KEY *PKCS11_find_key(PKCS11_CERT *);
 
-/* Find the corresponding key (if any)  pub <-> priv base on ID */
-extern PKCS11_KEY *PKCS11_find_key_from_key(PKCS11_KEY *);
-
 /* Get a list of all certificates associated with this token */
 extern int PKCS11_enumerate_certs(PKCS11_TOKEN *, PKCS11_CERT **, unsigned int *);
 
+/* Get a list of all certificates associated with this token and matching cert template */
+extern int PKCS11_enumerate_certs_ext(PKCS11_TOKEN *,
+	const PKCS11_CERT *, PKCS11_CERT **, unsigned int *);
+
+/* Remove the certificate from this token */
+extern int PKCS11_remove_certificate(PKCS11_CERT *);
+
 /* Set UI method to allow retrieving CKU_CONTEXT_SPECIFIC PINs interactively */
 extern int PKCS11_set_ui_method(PKCS11_CTX *ctx,
 	UI_METHOD *ui_method, void *ui_user_data);
@@ -294,7 +327,7 @@ extern int PKCS11_set_ui_method(PKCS11_CTX *ctx,
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_init_token(PKCS11_TOKEN * token, const char *pin,
+extern int PKCS11_init_token(PKCS11_TOKEN *token, const char *pin,
 	const char *label);
 
 /**
@@ -305,10 +338,10 @@ extern int PKCS11_init_token(PKCS11_TOKEN * token, const char *pin,
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_init_pin(PKCS11_TOKEN * token, const char *pin);
+extern int PKCS11_init_pin(PKCS11_TOKEN *token, const char *pin);
 
 /**
- * Change the user PIN on a token
+ * Change the currently used (either USER or SO) PIN on a token.
  *
  * @param slot slot returned by PKCS11_find_token()
  * @param old_pin old PIN value
@@ -316,7 +349,7 @@ extern int PKCS11_init_pin(PKCS11_TOKEN * token, const char *pin);
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_change_pin(PKCS11_SLOT * slot, const char *old_pin,
+extern int PKCS11_change_pin(PKCS11_SLOT *slot, const char *old_pin,
 	const char *new_pin);
 
 /**
@@ -325,12 +358,12 @@ extern int PKCS11_change_pin(PKCS11_SLOT * slot, const char *old_pin,
  * @param token token returned by PKCS11_find_token()
  * @param pk private key
  * @param label label for this key
- * @param id bytes to use as id value
- * @param id_len length of id value.
+ * @param id bytes to use as the id value
+ * @param id_len length of the id value
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_store_private_key(PKCS11_TOKEN * token, EVP_PKEY * pk, char *label, unsigned char *id, size_t id_len);
+extern int PKCS11_store_private_key(PKCS11_TOKEN *token, EVP_PKEY *pk, char *label, unsigned char *id, size_t id_len);
 
 /**
  * Store public key on a token
@@ -338,12 +371,12 @@ extern int PKCS11_store_private_key(PKCS11_TOKEN * token, EVP_PKEY * pk, char *l
  * @param token token returned by PKCS11_find_token()
  * @param pk private key
  * @param label label for this key
- * @param id bytes to use as id value
- * @param id_len length of id value.
+ * @param id bytes to use as the id value
+ * @param id_len length of the id value
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_store_public_key(PKCS11_TOKEN * token, EVP_PKEY * pk, char *label, unsigned char *id, size_t id_len);
+extern int PKCS11_store_public_key(PKCS11_TOKEN *token, EVP_PKEY *pk, char *label, unsigned char *id, size_t id_len);
 
 /**
  * Store certificate on a token
@@ -351,13 +384,13 @@ extern int PKCS11_store_public_key(PKCS11_TOKEN * token, EVP_PKEY * pk, char *la
  * @param token token returned by PKCS11_find_token()
  * @param x509 x509 certificate object
  * @param label label for this certificate
- * @param id bytes to use as id value
- * @param id_len length of id value.
+ * @param id bytes to use as the id value
+ * @param id_len length of the id value
  * @param ret_cert put new PKCS11_CERT object here
  * @retval 0 success
  * @retval -1 error
  */
-extern int PKCS11_store_certificate(PKCS11_TOKEN * token, X509 * x509,
+extern int PKCS11_store_certificate(PKCS11_TOKEN *token, X509 *x509,
 		char *label, unsigned char *id, size_t id_len,
 		PKCS11_CERT **ret_cert);
 
@@ -379,6 +412,8 @@ void *PKCS11_get_ec_key_method(void);
 ECDSA_METHOD *PKCS11_get_ecdsa_method(void);
 ECDH_METHOD *PKCS11_get_ecdh_method(void);
 #endif
+int PKCS11_pkey_meths(ENGINE *e, EVP_PKEY_METHOD **pmeth,
+		const int **nids, int nid);
 
 /**
  * Load PKCS11 error strings
@@ -388,77 +423,55 @@ ECDH_METHOD *PKCS11_get_ecdh_method(void);
  */
 extern void ERR_load_PKCS11_strings(void);
 
-#if defined(_LIBP11_INT_H)
-	/* Deprecated functions will no longer be exported in libp11 0.5.0 */
-	/* They are, however, used internally in OpenSSL method definitions */
-#define P11_DEPRECATED(msg)
-#elif defined(_MSC_VER)
-#define P11_DEPRECATED(msg) __declspec(deprecated(msg))
-#elif defined(__GNUC__)
-#if (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__) >= 40500
-	/* GCC >= 4.5.0 supports printing a message */
-#define P11_DEPRECATED(msg) __attribute__ ((deprecated(msg)))
-#else
-#define P11_DEPRECATED(msg) __attribute__ ((deprecated))
-#endif
-#elif defined(__clang__)
-#define P11_DEPRECATED(msg) __attribute__ ((deprecated(msg)))
-#else
-#define P11_DEPRECATED(msg)
-#endif
-
-#define P11_DEPRECATED_FUNC \
-	P11_DEPRECATED("This function will be removed in libp11 0.5.0")
-
 /*
- * These functions will be removed from libp11, because they partially
+ * The following functions are discouraged, because they partially
  * duplicate the functionality OpenSSL provides for EVP_PKEY objects
  */
 
 /**
- * Generate and store a private key on the token
+ * Generate a private key on the token
  *
  * @param token token returned by PKCS11_find_token()
- * @param algorithm EVP_PKEY_RSA
+ * @param algorithm IGNORED (still here for backward compatibility)
  * @param bits size of the modulus in bits
  * @param label label for this key
- * @param id bytes to use as id value
- * @param id_len length of id value.
+ * @param id bytes to use as the id value
+ * @param id_len length of the id value
  * @retval 0 success
  * @retval -1 error
  */
-P11_DEPRECATED_FUNC extern int PKCS11_generate_key(PKCS11_TOKEN * token,
+extern int PKCS11_generate_key(PKCS11_TOKEN *token,
 	int algorithm, unsigned int bits,
-	char *label, unsigned char* id, size_t id_len);
+	char *label, unsigned char *id, size_t id_len);
 
 /* Get the RSA key modulus size (in bytes) */
-P11_DEPRECATED_FUNC extern int PKCS11_get_key_size(PKCS11_KEY *);
+extern int PKCS11_get_key_size(PKCS11_KEY *);
 
 /* Get the RSA key modules as BIGNUM */
-P11_DEPRECATED_FUNC extern int PKCS11_get_key_modulus(PKCS11_KEY *, BIGNUM **);
+extern int PKCS11_get_key_modulus(PKCS11_KEY *, BIGNUM **);
 
 /* Get the RSA key public exponent as BIGNUM */
-P11_DEPRECATED_FUNC extern int PKCS11_get_key_exponent(PKCS11_KEY *, BIGNUM **);
+extern int PKCS11_get_key_exponent(PKCS11_KEY *, BIGNUM **);
 
 /* Sign with the EC private key */
-P11_DEPRECATED_FUNC extern int PKCS11_ecdsa_sign(
+extern int PKCS11_ecdsa_sign(
 	const unsigned char *m, unsigned int m_len,
-	unsigned char *sigret, unsigned int *siglen, PKCS11_KEY * key);
+	unsigned char *sigret, unsigned int *siglen, PKCS11_KEY *key);
 
 /* Sign with the RSA private key */
-P11_DEPRECATED_FUNC extern int PKCS11_sign(int type,
+extern int PKCS11_sign(int type,
 	const unsigned char *m, unsigned int m_len,
-	unsigned char *sigret, unsigned int *siglen, PKCS11_KEY * key);
+	unsigned char *sigret, unsigned int *siglen, PKCS11_KEY *key);
 
 /* This function has never been implemented */
-P11_DEPRECATED_FUNC extern int PKCS11_verify(int type,
+extern int PKCS11_verify(int type,
 	const unsigned char *m, unsigned int m_len,
-	unsigned char *signature, unsigned int siglen, PKCS11_KEY * key);
+	unsigned char *signature, unsigned int siglen, PKCS11_KEY *key);
 
 /* Encrypts data using the private key */
-P11_DEPRECATED_FUNC extern int PKCS11_private_encrypt(
+extern int PKCS11_private_encrypt(
 	int flen, const unsigned char *from,
-	unsigned char *to, PKCS11_KEY * rsa, int padding);
+	unsigned char *to, PKCS11_KEY *rsa, int padding);
 
 /**
  * Decrypts data using the private key
@@ -470,49 +483,90 @@ P11_DEPRECATED_FUNC extern int PKCS11_private_encrypt(
  * @param  padding  padding algorithm to be used
  * @return the length of the decrypted data or 0 if an error occurred
  */
-P11_DEPRECATED_FUNC extern int PKCS11_private_decrypt(
+extern int PKCS11_private_decrypt(
 	int flen, const unsigned char *from,
-	unsigned char *to, PKCS11_KEY * key, int padding); 
+	unsigned char *to, PKCS11_KEY *key, int padding);
 
-/*
- * Function and reason codes
- */
-#define PKCS11_F_PKCS11_CTX_LOAD		1
-#define PKCS11_F_PKCS11_ENUM_SLOTS		2
-#define PKCS11_F_PKCS11_CHECK_TOKEN		3
-#define PKCS11_F_PKCS11_OPEN_SESSION		4
-#define PKCS11_F_PKCS11_LOGIN			5
-#define PKCS11_F_PKCS11_ENUM_KEYS		6
-#define PKCS11_F_PKCS11_GET_KEY			7
-#define PKCS11_F_PKCS11_RSA_DECRYPT		8
-#define PKCS11_F_PKCS11_RSA_ENCRYPT		9
-#define PKCS11_F_PKCS11_RSA_SIGN		10
-#define PKCS11_F_PKCS11_RSA_VERIFY		11
-#define PKCS11_F_PKCS11_ENUM_CERTS		12
-#define PKCS11_F_PKCS11_INIT_TOKEN		13
-#define PKCS11_F_PKCS11_INIT_PIN		14
-#define PKCS11_F_PKCS11_LOGOUT			15
-#define PKCS11_F_PKCS11_STORE_PRIVATE_KEY	16
-#define PKCS11_F_PKCS11_GENERATE_KEY		17
-#define PKCS11_F_PKCS11_STORE_PUBLIC_KEY	18
-#define PKCS11_F_PKCS11_STORE_CERTIFICATE	19
-#define PKCS11_F_PKCS11_SEED_RANDOM		20
-#define PKCS11_F_PKCS11_GENERATE_RANDOM		21
-#define PKCS11_F_PKCS11_CHANGE_PIN		22
-#define PKCS11_F_PKCS11_GETATTR			40
-#define PKCS11_F_PKCS11_EC_KEY_SIGN			41
-#define PKCS11_F_PKCS11_EC_KEY_VERIFY		42
-#define PKCS11_F_PKCS11_GETSESSIONINFO		43
-#define PKCS11_F_PKCS11_EC_KEY_COMPUTE_KEY	44
+/* Function codes */
+# define CKR_F_PKCS11_CHANGE_PIN                          100
+# define CKR_F_PKCS11_CHECK_TOKEN                         101
+# define CKR_F_PKCS11_CTX_LOAD                            102
+# define CKR_F_PKCS11_ECDH_DERIVE                         103
+# define CKR_F_PKCS11_ECDSA_SIGN                          104
+# define CKR_F_PKCS11_ENUMERATE_SLOTS                     105
+# define CKR_F_PKCS11_FIND_CERTS                          106
+# define CKR_F_PKCS11_FIND_KEYS                           107
+# define CKR_F_PKCS11_GENERATE_RANDOM                     108
+# define CKR_F_PKCS11_GETATTR_ALLOC                       109
+# define CKR_F_PKCS11_GETATTR_BN                          110
+# define CKR_F_PKCS11_GETATTR_INT                         111
+# define CKR_F_PKCS11_INIT_PIN                            112
+# define CKR_F_PKCS11_INIT_SLOT                           113
+# define CKR_F_PKCS11_INIT_TOKEN                          114
+# define CKR_F_PKCS11_IS_LOGGED_IN                        115
+# define CKR_F_PKCS11_LOGIN                               116
+# define CKR_F_PKCS11_LOGOUT                              117
+# define CKR_F_PKCS11_NEXT_CERT                           118
+# define CKR_F_PKCS11_NEXT_KEY                            119
+# define CKR_F_PKCS11_OPEN_SESSION                        120
+# define CKR_F_PKCS11_PRIVATE_DECRYPT                     121
+# define CKR_F_PKCS11_PRIVATE_ENCRYPT                     122
+# define CKR_F_PKCS11_RELOAD_KEY                          123
+# define CKR_F_PKCS11_SEED_RANDOM                         125
+# define CKR_F_PKCS11_STORE_CERTIFICATE                   126
+# define CKR_F_PKCS11_STORE_KEY                           127
+# define CKR_F_PKCS11_REMOVE_KEY                          128
+# define CKR_F_PKCS11_REMOVE_CERTIFICATE                  129
+# define CKR_F_PKCS11_GENERATE_KEY                        130
+# define CKR_F_PKCS11_RELOAD_CERTIFICATE                  131
+# define CKR_F_PKCS11_GET_SESSION                         132
 
-#define PKCS11_ERR_BASE				1024
-#define PKCS11_LOAD_MODULE_ERROR		(PKCS11_ERR_BASE+1)
-#define PKCS11_MODULE_LOADED_ERROR		(PKCS11_ERR_BASE+2)
-#define PKCS11_SYMBOL_NOT_FOUND_ERROR		(PKCS11_ERR_BASE+3)
-#define PKCS11_NOT_SUPPORTED			(PKCS11_ERR_BASE+4)
-#define PKCS11_NO_SESSION			(PKCS11_ERR_BASE+5)
-#define PKCS11_KEYGEN_FAILED			(PKCS11_ERR_BASE+6)
-#define PKCS11_UI_FAILED			(PKCS11_ERR_BASE+7)
+/* Backward compatibility of error function codes */
+#define PKCS11_F_PKCS11_CHANGE_PIN CKR_F_PKCS11_CHANGE_PIN
+#define PKCS11_F_PKCS11_CHECK_TOKEN CKR_F_PKCS11_CHECK_TOKEN
+#define PKCS11_F_PKCS11_CTX_LOAD CKR_F_PKCS11_CTX_LOAD
+#define PKCS11_F_PKCS11_ECDH_DERIVE CKR_F_PKCS11_ECDH_DERIVE
+#define PKCS11_F_PKCS11_ECDSA_SIGN CKR_F_PKCS11_ECDSA_SIGN
+#define PKCS11_F_PKCS11_ENUMERATE_SLOTS CKR_F_PKCS11_ENUMERATE_SLOTS
+#define PKCS11_F_PKCS11_FIND_CERTS CKR_F_PKCS11_FIND_CERTS
+#define PKCS11_F_PKCS11_FIND_KEYS CKR_F_PKCS11_FIND_KEYS
+#define PKCS11_F_PKCS11_GENERATE_RANDOM CKR_F_PKCS11_GENERATE_RANDOM
+#define PKCS11_F_PKCS11_GETATTR_ALLOC CKR_F_PKCS11_GETATTR_ALLOC
+#define PKCS11_F_PKCS11_GETATTR_BN CKR_F_PKCS11_GETATTR_BN
+#define PKCS11_F_PKCS11_GETATTR_INT CKR_F_PKCS11_GETATTR_INT
+#define PKCS11_F_PKCS11_INIT_PIN CKR_F_PKCS11_INIT_PIN
+#define PKCS11_F_PKCS11_INIT_SLOT CKR_F_PKCS11_INIT_SLOT
+#define PKCS11_F_PKCS11_INIT_TOKEN CKR_F_PKCS11_INIT_TOKEN
+#define PKCS11_F_PKCS11_IS_LOGGED_IN CKR_F_PKCS11_IS_LOGGED_IN
+#define PKCS11_F_PKCS11_LOGIN CKR_F_PKCS11_LOGIN
+#define PKCS11_F_PKCS11_LOGOUT CKR_F_PKCS11_LOGOUT
+#define PKCS11_F_PKCS11_NEXT_CERT CKR_F_PKCS11_NEXT_CERT
+#define PKCS11_F_PKCS11_NEXT_KEY CKR_F_PKCS11_NEXT_KEY
+#define PKCS11_F_PKCS11_OPEN_SESSION CKR_F_PKCS11_OPEN_SESSION
+#define PKCS11_F_PKCS11_PRIVATE_DECRYPT CKR_F_PKCS11_PRIVATE_DECRYPT
+#define PKCS11_F_PKCS11_PRIVATE_ENCRYPT CKR_F_PKCS11_PRIVATE_ENCRYPT
+#define PKCS11_F_PKCS11_RELOAD_KEY CKR_F_PKCS11_RELOAD_KEY
+#define PKCS11_F_PKCS11_SEED_RANDOM CKR_F_PKCS11_SEED_RANDOM
+#define PKCS11_F_PKCS11_STORE_CERTIFICATE CKR_F_PKCS11_STORE_CERTIFICATE
+#define PKCS11_F_PKCS11_STORE_KEY CKR_F_PKCS11_STORE_KEY
+#define PKCS11_F_PKCS11_REMOVE_KEY CKR_F_PKCS11_REMOVE_KEY
+#define PKCS11_F_PKCS11_REMOVE_CERTIFICATE CKR_F_PKCS11_REMOVE_CERTIFICATE
+#define PKCS11_F_PKCS11_GENERATE_KEY CKR_F_PKCS11_GENERATE_KEY
+
+/* Backward compatibility of error reason codes */
+#define PKCS11_LOAD_MODULE_ERROR                          P11_R_LOAD_MODULE_ERROR
+#define PKCS11_MODULE_LOADED_ERROR                        -1
+#define PKCS11_SYMBOL_NOT_FOUND_ERROR                     -1
+#define PKCS11_NOT_SUPPORTED                              P11_R_NOT_SUPPORTED
+#define PKCS11_NO_SESSION                                 P11_R_NO_SESSION
+#define PKCS11_KEYGEN_FAILED                              P11_R_KEYGEN_FAILED
+#define PKCS11_UI_FAILED                                  P11_R_UI_FAILED
+
+/* Backward compatibility emulation of the ERR_LIB_PKCS11 constant.
+ * We currently use two separate variables for library error codes:
+ * one for imported PKCS#11 module errors, and one for our own libp11 errors.
+ * We return the value for PKCS#11, as it is more likely to be needed. */
+#define ERR_LIB_PKCS11 (ERR_get_CKR_code())
 
 #ifdef __cplusplus
 }
