@@ -123,20 +123,20 @@ unsigned int pkcs11_addattr(PKCS11_TEMPLATE *tmpl, int type, void *data, size_t
 
 void pkcs11_addattr_bool(PKCS11_TEMPLATE *tmpl, int type, int value)
 {
-	static CK_BBOOL true = CK_TRUE;
-	static CK_BBOOL false = CK_FALSE;
-	pkcs11_addattr(tmpl, type, value ? &true : &false, sizeof(CK_BBOOL));
+	static CK_BBOOL _true = CK_TRUE;
+	static CK_BBOOL _false = CK_FALSE;
+	pkcs11_addattr(tmpl, type, value ? &_true : &_false, sizeof(CK_BBOOL));
 }
 
 void pkcs11_addattr_s(PKCS11_TEMPLATE *tmpl, int type, const char *s)
 {
-	pkcs11_addattr(tmpl, type, (void*) s, s ? strlen(s) : 0);
+	pkcs11_addattr(tmpl, type, (void *)s, s ? strlen(s) : 0);
 }
 
 void pkcs11_addattr_bn(PKCS11_TEMPLATE *tmpl, int type, const BIGNUM *bn)
 {
 	int n = BN_num_bytes(bn);
-	uint8_t *buf = OPENSSL_malloc(n);
+	unsigned char *buf = OPENSSL_malloc(n);
 	unsigned int i;
 
 	if (buf && BN_bn2bin(bn, buf) == n) {
@@ -162,9 +162,11 @@ void pkcs11_addattr_obj(PKCS11_TEMPLATE *tmpl, int type, pkcs11_i2d_fn enc, void
 
 void pkcs11_zap_attrs(PKCS11_TEMPLATE *tmpl)
 {
+	unsigned int i;
+
 	if (!tmpl->allocated)
 		return;
-	for (unsigned i = 0; i < 32; i++) {
+	for (i = 0; i < 32; i++) {
 		if (tmpl->allocated & (1<<i))
 			OPENSSL_free(tmpl->attrs[i].pValue);
 	}
