@@ -26,23 +26,34 @@
 #include "libp11-int.h"
 #include <string.h>
 
-static int pkcs11_find_certs(PKCS11_SLOT_private *, CK_SESSION_HANDLE);
+static int pkcs11_find_certs(PKCS11_SLOT_private *, PKCS11_TEMPLATE *, CK_SESSION_HANDLE);
 static int pkcs11_next_cert(PKCS11_CTX_private *, PKCS11_SLOT_private *, CK_SESSION_HANDLE);
 static int pkcs11_init_cert(PKCS11_SLOT_private *token, CK_SESSION_HANDLE session,
 	CK_OBJECT_HANDLE o, PKCS11_CERT **);
 
 /*
- * Enumerate all certs on the card
+ * Enumerate all certs matching with cert_template on the card
  */
-int pkcs11_enumerate_certs(PKCS11_SLOT_private *slot, PKCS11_CERT **certp, unsigned int *countp)
+int pkcs11_enumerate_certs(PKCS11_SLOT_private *slot, const PKCS11_CERT *cert_template, PKCS11_CERT **certp, unsigned int *countp)
 {
 	CK_SESSION_HANDLE session;
 	int rv;
+	PKCS11_TEMPLATE tmpl = {0};
+	CK_OBJECT_CLASS object_class = CKO_CERTIFICATE;
+	pkcs11_addattr_var(&tmpl, CKA_CLASS, object_class);
+
+	if (cert_template) {
+		if (cert_template->id_len)
+			pkcs11_addattr(&tmpl, CKA_ID, cert_template->id, cert_template->id_len);
+
+		if (cert_template->label)
+			pkcs11_addattr_s(&tmpl, CKA_LABEL, cert_template->label);
+	}
 
 	if (pkcs11_get_session(slot, 0, &session))
 		return -1;
 
-	rv = pkcs11_find_certs(slot, session);
+	rv = pkcs11_find_certs(slot, &tmpl, session);
 	pkcs11_put_session(slot, session);
 	if (rv < 0) {
 		pkcs11_destroy_certs(slot);
@@ -62,10 +73,12 @@ int pkcs11_enumerate_certs(PKCS11_SLOT_private *slot, PKCS11_CERT **certp, unsig
 PKCS11_CERT *pkcs11_find_certificate(PKCS11_OBJECT_private *key)
 {
 	PKCS11_OBJECT_private *cpriv;
-	PKCS11_CERT *cert;
+	PKCS11_CERT *cert, cert_template = {0};
 	unsigned int n, count;
 
-	if (pkcs11_enumerate_certs(key->slot, &cert, &count))
+	cert_template.id = key->id;
+	cert_template.id_len = key->id_len;
+	if (pkcs11_enumerate_certs(key->slot, &cert_template, &cert, &count))
 		return NULL;
 	for (n = 0; n < count; n++, cert++) {
 		cpriv = PRIVCERT(cert);
@@ -77,20 +90,15 @@ PKCS11_CERT *pkcs11_find_certificate(PKCS11_OBJECT_private *key)
 }
 
 /*
- * Find all certs of a given type (public or private)
+ * Find all certs of a given type (public or private) and matching template
  */
-static int pkcs11_find_certs(PKCS11_SLOT_private *slot, CK_SESSION_HANDLE session)
+static int pkcs11_find_certs(PKCS11_SLOT_private *slot, PKCS11_TEMPLATE *tmpl, CK_SESSION_HANDLE session)
 {
 	PKCS11_CTX_private *ctx = slot->ctx;
-	CK_OBJECT_CLASS cert_search_class;
-	CK_ATTRIBUTE cert_search_attrs[] = {
-		{CKA_CLASS, &cert_search_class, sizeof(cert_search_class)},
-	};
 	int rv, res = -1;
 
 	/* Tell the PKCS11 lib to enumerate all matching objects */
-	cert_search_class = CKO_CERTIFICATE;
-	rv = CRYPTOKI_call(ctx, C_FindObjectsInit(session, cert_search_attrs, 1));
+	rv = CRYPTOKI_call(ctx, C_FindObjectsInit(session, tmpl->attrs, tmpl->nattr));
 	CRYPTOKI_checkerr(CKR_F_PKCS11_FIND_CERTS, rv);
 
 	do {
@@ -123,7 +131,7 @@ static int pkcs11_next_cert(PKCS11_CTX_private *ctx, PKCS11_SLOT_private *slot,
 }
 
 static int pkcs11_init_cert(PKCS11_SLOT_private *slot, CK_SESSION_HANDLE session,
-	CK_OBJECT_HANDLE object, PKCS11_CERT ** ret)
+	CK_OBJECT_HANDLE object, PKCS11_CERT **ret)
 {
 	PKCS11_OBJECT_private *cpriv;
 	PKCS11_CERT *cert, *tmp;
@@ -194,7 +202,7 @@ int pkcs11_store_certificate(PKCS11_SLOT_private *slot, X509 *x509, char *label,
 	int rv, r = -1;
 	int signature_nid;
 	int evp_md_nid = NID_sha1;
-	const EVP_MD* evp_md;
+	const EVP_MD *evp_md;
 	unsigned char md[EVP_MAX_MD_SIZE];
 	unsigned int md_len;
 	PKCS11_TEMPLATE tmpl = {0};
@@ -216,54 +224,61 @@ int pkcs11_store_certificate(PKCS11_SLOT_private *slot, X509 *x509, char *label,
 		(pkcs11_i2d_fn)i2d_X509_NAME, X509_get_issuer_name(x509));
 
 	/* Get digest algorithm from x509 certificate */
+	/* CKA_NAME_HASH_ALGORITHM was added in Cryptoki 2.30;
+	 * older versions of PKCS#11 modules should not touch
+	 * this attribute or any other related attributes */
+	if (ctx->cryptoki_version.major > 2 ||
+		(ctx->cryptoki_version.major == 2 && ctx->cryptoki_version.minor >= 30)) {
+		/* Get digest algorithm from x509 certificate */
 #if OPENSSL_VERSION_NUMBER >= 0x10002000L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER >= 0x3050000fL )
-	signature_nid = X509_get_signature_nid(x509);
+		signature_nid = X509_get_signature_nid(x509);
 #else
-	signature_nid = OBJ_obj2nid(x509->sig_alg->algorithm);
+		signature_nid = OBJ_obj2nid(x509->sig_alg->algorithm);
 #endif
-	OBJ_find_sigid_algs(signature_nid, &evp_md_nid, NULL);
-	switch (evp_md_nid) {
-	default:
-		evp_md_nid = NID_sha1;
-		/* fall through */
-	case NID_sha1:
-		ckm_md = CKM_SHA_1;
-		break;
-	case NID_sha224:
-		ckm_md = CKM_SHA224;
-		break;
-	case NID_sha256:
-		ckm_md = CKM_SHA256;
-		break;
-	case NID_sha512:
-		ckm_md = CKM_SHA512;
-		break;
-	case NID_sha384:
-		ckm_md = CKM_SHA384;
-		break;
-#if !defined(LIBRESSL_VERSION_NUMBER)
-	case NID_sha3_224:
-		ckm_md = CKM_SHA3_224;
-		break;
-	case NID_sha3_256:
-		ckm_md = CKM_SHA3_256;
-		break;
-	case NID_sha3_384:
-		ckm_md = CKM_SHA3_384;
-		break;
-	case NID_sha3_512:
-		ckm_md = CKM_SHA3_512;
-		break;
+		OBJ_find_sigid_algs(signature_nid, &evp_md_nid, NULL);
+		switch (evp_md_nid) {
+		default:
+			evp_md_nid = NID_sha1;
+			/* fall through */
+		case NID_sha1:
+			ckm_md = CKM_SHA_1;
+			break;
+		case NID_sha224:
+			ckm_md = CKM_SHA224;
+			break;
+		case NID_sha256:
+			ckm_md = CKM_SHA256;
+			break;
+		case NID_sha512:
+			ckm_md = CKM_SHA512;
+			break;
+		case NID_sha384:
+			ckm_md = CKM_SHA384;
+			break;
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L && !defined(LIBRESSL_VERSION_NUMBER)
+		case NID_sha3_224:
+			ckm_md = CKM_SHA3_224;
+			break;
+		case NID_sha3_256:
+			ckm_md = CKM_SHA3_256;
+			break;
+		case NID_sha3_384:
+			ckm_md = CKM_SHA3_384;
+			break;
+		case NID_sha3_512:
+			ckm_md = CKM_SHA3_512;
+			break;
 #endif
+		}
+
+		evp_md = EVP_get_digestbynid(evp_md_nid);
+
+		/* Set hash algorithm; default is SHA-1 */
+		pkcs11_addattr_var(&tmpl, CKA_NAME_HASH_ALGORITHM, ckm_md);
+		if (X509_pubkey_digest(x509, evp_md, md, &md_len))
+			pkcs11_addattr(&tmpl, CKA_HASH_OF_SUBJECT_PUBLIC_KEY, md, md_len);
 	}
 
-	evp_md = EVP_get_digestbynid(evp_md_nid);
-
-	/* Set hash algorithm; default is SHA-1 */
-	pkcs11_addattr_var(&tmpl, CKA_NAME_HASH_ALGORITHM, ckm_md);
-	if (X509_pubkey_digest(x509,evp_md,md,&md_len))
-		pkcs11_addattr(&tmpl, CKA_HASH_OF_SUBJECT_PUBLIC_KEY, md, md_len);
-
 	pkcs11_addattr_obj(&tmpl, CKA_VALUE, (pkcs11_i2d_fn)i2d_X509, x509);
 	if (label)
 		pkcs11_addattr_s(&tmpl, CKA_LABEL, label);
