@@ -38,6 +38,19 @@
 #include <openssl/ecdh.h>
 #endif
 
+#if OPENSSL_VERSION_NUMBER >= 0x10100002L
+#ifndef OPENSSL_NO_EC
+static EC_KEY_METHOD *pkcs11_ec_key_method = NULL;
+#endif /* OPENSSL_NO_EC */
+#else /* OPENSSL_VERSION_NUMBER */
+#ifndef OPENSSL_NO_ECDSA
+static ECDSA_METHOD *pkcs11_ecdsa_method = NULL;
+#endif /* OPENSSL_NO_ECDSA */
+#ifndef OPENSSL_NO_ECDH
+static ECDH_METHOD *pkcs11_ecdh_method = NULL;
+#endif /* OPENSSL_NO_ECDH */
+#endif /* OPENSSL_VERSION_NUMBER */
+
 #ifndef OPENSSL_NO_EC
 
 #if OPENSSL_VERSION_NUMBER >= 0x10100004L && !defined(LIBRESSL_VERSION_NUMBER)
@@ -50,6 +63,7 @@ typedef int (*compute_key_fn)(void *, size_t,
 #endif
 static compute_key_fn ossl_ecdh_compute_key;
 static void (*ossl_ec_finish)(EC_KEY *);
+static int (*ossl_ec_copy)(EC_KEY *, const EC_KEY *);
 
 static int ec_ex_index = 0;
 
@@ -60,6 +74,8 @@ typedef ECDSA_SIG *(*sign_sig_fn)(const unsigned char *, int,
 
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
 
+#define EC_KEY_METHOD_DYNAMIC   1
+
 /* ecdsa_method maintains unchanged layout between 0.9.8 and 1.0.2 */
 
 /* Data pointers and function pointers may have different sizes on some
@@ -170,8 +186,6 @@ static void alloc_ec_ex_index()
 	}
 }
 
-#if 0
-/* TODO: Free the indexes on unload */
 static void free_ec_ex_index()
 {
 	if (ec_ex_index > 0) {
@@ -182,7 +196,6 @@ static void free_ec_ex_index()
 		ec_ex_index = 0;
 	}
 }
-#endif
 
 /********** EVP_PKEY retrieval */
 
@@ -247,10 +260,13 @@ static int pkcs11_get_point(EC_KEY *ec, PKCS11_OBJECT_private *key, CK_SESSION_H
 	ASN1_OCTET_STRING *os;
 	int rv = -1;
 
+	if (!key)
+		return -1;
+
 	if (key->x509 && pkcs11_get_point_x509(ec, key->x509) == 0)
 		return 0;
 
-	if (!key || pkcs11_getattr_alloc(key->slot->ctx, session, key->object,
+	if (pkcs11_getattr_alloc(key->slot->ctx, session, key->object,
 			CKA_EC_POINT, &point, &point_len))
 		return -1;
 
@@ -335,7 +351,7 @@ PKCS11_OBJECT_private *pkcs11_get_ex_data_ec(const EC_KEY *ec)
 #endif
 }
 
-static void pkcs11_set_ex_data_ec(EC_KEY *ec, PKCS11_OBJECT_private *key)
+void pkcs11_set_ex_data_ec(EC_KEY *ec, PKCS11_OBJECT_private *key)
 {
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
 	EC_KEY_set_ex_data(ec, ec_ex_index, key);
@@ -374,13 +390,16 @@ static EVP_PKEY *pkcs11_get_evp_key_ec(PKCS11_OBJECT_private *key)
 		ECDSA_set_method(ec, PKCS11_get_ecdsa_method());
 		ECDH_set_method(ec, PKCS11_get_ecdh_method());
 #endif
+		/* This creates a new EC_KEY object which requires its own key object reference */
+		key = pkcs11_object_ref(key);
+		pkcs11_set_ex_data_ec(ec, key);
 	}
 	/* TODO: Retrieve the ECDSA private key object attributes instead,
 	 * unless the key has the "sensitive" attribute set */
 
-	pkcs11_set_ex_data_ec(ec, key);
 	EVP_PKEY_set1_EC_KEY(pk, ec); /* Also increments the ec ref count */
 	EC_KEY_free(ec); /* Drops our reference to it */
+
 	return pk;
 }
 
@@ -570,36 +589,36 @@ static int pkcs11_ecdh_derive(unsigned char **out, size_t *outlen,
 	CK_MECHANISM mechanism;
 	int rv;
 
-	CK_BBOOL true = TRUE;
-	CK_BBOOL false = FALSE;
+	CK_BBOOL _true = TRUE;
+	CK_BBOOL _false = FALSE;
 	CK_OBJECT_HANDLE newkey = CK_INVALID_HANDLE;
-	CK_OBJECT_CLASS newkey_class= CKO_SECRET_KEY;
+	CK_OBJECT_CLASS newkey_class = CKO_SECRET_KEY;
 	CK_KEY_TYPE newkey_type = CKK_GENERIC_SECRET;
 	CK_ULONG newkey_len = key_len;
 	CK_OBJECT_HANDLE *tmpnewkey = (CK_OBJECT_HANDLE *)outnewkey;
 	CK_ATTRIBUTE newkey_template[] = {
-		{CKA_TOKEN, &false, sizeof(false)}, /* session only object */
+		{CKA_TOKEN, &_false, sizeof(_false)}, /* session only object */
 		{CKA_CLASS, &newkey_class, sizeof(newkey_class)},
 		{CKA_KEY_TYPE, &newkey_type, sizeof(newkey_type)},
 		{CKA_VALUE_LEN, &newkey_len, sizeof(newkey_len)},
-		{CKA_SENSITIVE, &false, sizeof(false) },
-		{CKA_EXTRACTABLE, &true, sizeof(true) },
-		{CKA_ENCRYPT, &true, sizeof(true)},
-		{CKA_DECRYPT, &true, sizeof(true)}
+		{CKA_SENSITIVE, &_false, sizeof(_false) },
+		{CKA_EXTRACTABLE, &_true, sizeof(_true) },
+		{CKA_ENCRYPT, &_true, sizeof(_true)},
+		{CKA_DECRYPT, &_true, sizeof(_true)}
 	};
 
 	memset(&mechanism, 0, sizeof(mechanism));
-	mechanism.mechanism  = ecdh_mechanism;
-	mechanism.pParameter =  (void*)ec_params;
+	mechanism.mechanism = ecdh_mechanism;
+	mechanism.pParameter = (void *)ec_params;
 	switch (ecdh_mechanism) {
 		case CKM_ECDH1_DERIVE:
 		case CKM_ECDH1_COFACTOR_DERIVE:
-			mechanism.ulParameterLen  = sizeof(CK_ECDH1_DERIVE_PARAMS);
+			mechanism.ulParameterLen = sizeof(CK_ECDH1_DERIVE_PARAMS);
 			break;
 #if 0
 		/* TODO */
 		case CK_ECMQV_DERIVE_PARAMS:
-			mechanism.ulParameterLen  = sizeof(CK_ECMQV_DERIVE_PARAMS);
+			mechanism.ulParameterLen = sizeof(CK_ECMQV_DERIVE_PARAMS);
 			break;
 #endif
 		default:
@@ -681,6 +700,27 @@ static int pkcs11_ec_ckey(unsigned char **out, size_t *outlen,
 	return 1;
 }
 
+/* Without this, the EC_KEY objects share the same PKCS11_OBJECT_private
+ * object in ex_data and when one of them is freed, the following frees
+ * result in crashes.
+ * We need to increase the reference to the private object.
+ */
+static int pkcs11_ec_copy(EC_KEY *dest, const EC_KEY *src)
+{
+	PKCS11_OBJECT_private *srckey = NULL;
+	PKCS11_OBJECT_private *destkey = NULL;
+
+	srckey = pkcs11_get_ex_data_ec(src);
+	destkey = pkcs11_object_ref(srckey);
+
+	pkcs11_set_ex_data_ec(dest, destkey);
+
+	if (ossl_ec_copy)
+		ossl_ec_copy(dest, src);
+
+	return 1;
+}
+
 #else
 
 /**
@@ -738,31 +778,43 @@ static int pkcs11_ec_ckey(void *out, size_t outlen,
 
 EC_KEY_METHOD *PKCS11_get_ec_key_method(void)
 {
-	static EC_KEY_METHOD *ops = NULL;
 	int (*orig_init)(EC_KEY *);
-	int (*orig_copy)(EC_KEY *, const EC_KEY *);
 	int (*orig_set_group)(EC_KEY *, const EC_GROUP *);
 	int (*orig_set_private)(EC_KEY *, const BIGNUM *);
 	int (*orig_set_public)(EC_KEY *, const EC_POINT *);
 	int (*orig_sign)(int, const unsigned char *, int, unsigned char *,
 		unsigned int *, const BIGNUM *, const BIGNUM *, EC_KEY *) = NULL;
 
-	alloc_ec_ex_index();
-	if (!ops) {
-		ops = EC_KEY_METHOD_new((EC_KEY_METHOD *)EC_KEY_OpenSSL());
-		EC_KEY_METHOD_get_init(ops, &orig_init, &ossl_ec_finish, &orig_copy,
+	if (!pkcs11_ec_key_method) {
+		alloc_ec_ex_index();
+		pkcs11_ec_key_method = EC_KEY_METHOD_new((EC_KEY_METHOD *)EC_KEY_OpenSSL());
+		EC_KEY_METHOD_get_init(pkcs11_ec_key_method, &orig_init, &ossl_ec_finish, &ossl_ec_copy,
 			&orig_set_group, &orig_set_private, &orig_set_public);
-		EC_KEY_METHOD_set_init(ops, orig_init, pkcs11_ec_finish, orig_copy,
+		EC_KEY_METHOD_set_init(pkcs11_ec_key_method, orig_init, pkcs11_ec_finish, pkcs11_ec_copy,
 			orig_set_group, orig_set_private, orig_set_public);
-		EC_KEY_METHOD_get_sign(ops, &orig_sign, NULL, NULL);
-		EC_KEY_METHOD_set_sign(ops, orig_sign, NULL, pkcs11_ecdsa_sign_sig);
-		EC_KEY_METHOD_get_compute_key(ops, &ossl_ecdh_compute_key);
-		EC_KEY_METHOD_set_compute_key(ops, pkcs11_ec_ckey);
+		EC_KEY_METHOD_get_sign(pkcs11_ec_key_method, &orig_sign, NULL, NULL);
+		EC_KEY_METHOD_set_sign(pkcs11_ec_key_method, orig_sign, NULL, pkcs11_ecdsa_sign_sig);
+		EC_KEY_METHOD_get_compute_key(pkcs11_ec_key_method, &ossl_ecdh_compute_key);
+		EC_KEY_METHOD_set_compute_key(pkcs11_ec_key_method, pkcs11_ec_ckey);
 	}
-	return ops;
+	return pkcs11_ec_key_method;
 }
 
-/* define old way to keep old engines working without ECDSA */
+void pkcs11_ec_key_method_free(void)
+{
+	if (pkcs11_ec_key_method) {
+		free_ec_ex_index();
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+		if (meth->pkcs11_ecdh_method & EC_KEY_METHOD_DYNAMIC)
+			OPENSSL_free(pkcs11_ec_key_method);
+#else
+		EC_KEY_METHOD_free(pkcs11_ec_key_method);
+#endif
+		pkcs11_ec_key_method = NULL;
+	}
+}
+
+/* define old way to keep old applications/engines working without ECDSA */
 void *PKCS11_get_ecdsa_method(void)
 {
 	return NULL;
@@ -783,27 +835,51 @@ void *PKCS11_get_ec_key_method(void)
 
 ECDSA_METHOD *PKCS11_get_ecdsa_method(void)
 {
-	static ECDSA_METHOD *ops = NULL;
-
-	if (!ops) {
+	if (!pkcs11_ecdsa_method) {
 		alloc_ec_ex_index();
-		ops = ECDSA_METHOD_new((ECDSA_METHOD *)ECDSA_OpenSSL());
-		ECDSA_METHOD_set_sign(ops, pkcs11_ecdsa_sign_sig);
+		pkcs11_ecdsa_method = ECDSA_METHOD_new((ECDSA_METHOD *)ECDSA_OpenSSL());
+		ECDSA_METHOD_set_sign(pkcs11_ecdsa_method, pkcs11_ecdsa_sign_sig);
+	}
+	return pkcs11_ecdsa_method;
+}
+
+void pkcs11_ecdsa_method_free(void)
+{
+	if (pkcs11_ecdsa_method) {
+		free_ec_ex_index();
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+		if (pkcs11_ecdsa_method->flags & EC_KEY_METHOD_DYNAMIC)
+			OPENSSL_free(pkcs11_ecdsa_method);
+#else
+		EC_KEY_METHOD_free(pkcs11_ecdsa_method);
+#endif
+		pkcs11_ecdsa_method = NULL;
 	}
-	return ops;
 }
 
 ECDH_METHOD *PKCS11_get_ecdh_method(void)
 {
-	static ECDH_METHOD *ops = NULL;
-
-	if (!ops) {
+	if (!pkcs11_ecdh_method) {
 		alloc_ec_ex_index();
-		ops = ECDH_METHOD_new((ECDH_METHOD *)ECDH_OpenSSL());
-		ECDH_METHOD_get_compute_key(ops, &ossl_ecdh_compute_key);
-		ECDH_METHOD_set_compute_key(ops, pkcs11_ec_ckey);
+		pkcs11_ecdh_method = ECDH_METHOD_new((ECDH_METHOD *)ECDH_OpenSSL());
+		ECDH_METHOD_get_compute_key(pkcs11_ecdh_method, &ossl_ecdh_compute_key);
+		ECDH_METHOD_set_compute_key(pkcs11_ecdh_method, pkcs11_ec_ckey);
+	}
+	return pkcs11_ecdh_method;
+}
+
+void pkcs11_ecdh_method_free(void)
+{
+	if (pkcs11_ecdh_method) {
+		free_ec_ex_index();
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+		if (pkcs11_ecdh_method->flags & EC_KEY_METHOD_DYNAMIC)
+			OPENSSL_free(pkcs11_ecdh_method);
+#else
+		EC_KEY_METHOD_free(pkcs11_ecdh_method);
+#endif
+		pkcs11_ecdh_method = NULL;
 	}
-	return ops;
 }
 
 #endif /* OPENSSL_VERSION_NUMBER */
