@@ -1,6 +1,6 @@
 /* libp11, a simple layer on to of PKCS#11 API
  * Copyright (C) 2005 Olaf Kirch <okir@lst.de>
- * Copyright (C) 2016-2018 Michał Trojnara <Michal.Trojnara@stunnel.org>
+ * Copyright (C) 2016-2024 Michał Trojnara <Michal.Trojnara@stunnel.org>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -27,9 +27,9 @@
 #endif
 
 /* The maximum length of PIN */
-#define MAX_PIN_LENGTH   32
+#define MAX_PIN_LENGTH   256
 
-static int pkcs11_find_keys(PKCS11_SLOT_private *, CK_SESSION_HANDLE, unsigned int);
+static int pkcs11_find_keys(PKCS11_SLOT_private *, CK_SESSION_HANDLE, unsigned int, PKCS11_TEMPLATE *);
 static int pkcs11_next_key(PKCS11_CTX_private *ctx, PKCS11_SLOT_private *,
 	CK_SESSION_HANDLE session, CK_OBJECT_CLASS type);
 static int pkcs11_init_key(PKCS11_SLOT_private *, CK_SESSION_HANDLE session,
@@ -115,6 +115,8 @@ PKCS11_OBJECT_private *pkcs11_object_from_handle(PKCS11_SLOT_private *slot,
 		return NULL;
 
 	memset(obj, 0, sizeof(*obj));
+	obj->refcnt = 1;
+	pthread_mutex_init(&obj->lock, 0);
 	obj->object_class = object_class;
 	obj->object = object;
 	obj->slot = pkcs11_slot_ref(slot);
@@ -149,7 +151,8 @@ PKCS11_OBJECT_private *pkcs11_object_from_handle(PKCS11_SLOT_private *slot,
 PKCS11_OBJECT_private *pkcs11_object_from_template(PKCS11_SLOT_private *slot,
 	CK_SESSION_HANDLE session, PKCS11_TEMPLATE *tmpl)
 {
-	PKCS11_OBJECT_private *obj;
+	PKCS11_OBJECT_private *obj = NULL;
+	CK_OBJECT_HANDLE object_handle;
 	int release = 0;
 
 	if (session == CK_INVALID_HANDLE) {
@@ -158,8 +161,9 @@ PKCS11_OBJECT_private *pkcs11_object_from_template(PKCS11_SLOT_private *slot,
 		release = 1;
 	}
 
-	obj = pkcs11_object_from_handle(slot, session,
-		pkcs11_handle_from_template(slot, session, tmpl));
+	object_handle = pkcs11_handle_from_template(slot, session, tmpl);
+	if (object_handle)
+		obj = pkcs11_object_from_handle(slot, session, object_handle);
 
 	if (release)
 		pkcs11_put_session(slot, session);
@@ -178,17 +182,15 @@ PKCS11_OBJECT_private *pkcs11_object_from_object(PKCS11_OBJECT_private *obj,
 
 void pkcs11_object_free(PKCS11_OBJECT_private *obj)
 {
-	if (obj->evp_key) {
-		/* When the EVP object is reference count goes to zero,
-		 * it will call this function again. */
-		EVP_PKEY *pkey = obj->evp_key;
-		obj->evp_key = NULL;
-		EVP_PKEY_free(pkey);
+	if (!obj)
+		return;
+
+	if (pkcs11_atomic_add(&obj->refcnt, -1, &obj->lock) != 0)
 		return;
-	}
 	pkcs11_slot_unref(obj->slot);
 	X509_free(obj->x509);
 	OPENSSL_free(obj->label);
+	pthread_mutex_destroy(&obj->lock);
 	OPENSSL_free(obj);
 }
 
@@ -208,10 +210,14 @@ int pkcs11_set_ui_method(PKCS11_CTX_private *ctx,
  */
 PKCS11_KEY *pkcs11_find_key(PKCS11_OBJECT_private *cert)
 {
-	PKCS11_KEY *keys;
+	PKCS11_KEY *keys, key_template = {0};
 	unsigned int n, count;
+	key_template.isPrivate = 1;
 
-	if (pkcs11_enumerate_keys(cert->slot, CKO_PRIVATE_KEY, &keys, &count))
+	key_template.id = cert->id;
+	key_template.id_len = cert->id_len;
+
+	if (pkcs11_enumerate_keys(cert->slot, CKO_PRIVATE_KEY, &key_template, &keys, &count))
 		return NULL;
 	for (n = 0; n < count; n++) {
 		PKCS11_OBJECT_private *kpriv = PRIVKEY(&keys[n]);
@@ -253,7 +259,7 @@ int pkcs11_reload_object(PKCS11_OBJECT_private *obj)
  * Generate a key pair directly on token
  */
 int pkcs11_generate_key(PKCS11_SLOT_private *slot, int algorithm, unsigned int bits,
-		char *label, unsigned char* id, size_t id_len) {
+		char *label, unsigned char *id, size_t id_len) {
 
 	PKCS11_CTX_private *ctx = slot->ctx;
 	CK_SESSION_HANDLE session;
@@ -334,7 +340,7 @@ int pkcs11_store_public_key(PKCS11_SLOT_private *slot, EVP_PKEY *pk,
  */
 static int pkcs11_store_key(PKCS11_SLOT_private *slot, EVP_PKEY *pk,
 		CK_OBJECT_CLASS type, char *label, unsigned char *id, size_t id_len,
-		PKCS11_KEY ** ret_key)
+		PKCS11_KEY **ret_key)
 {
 	PKCS11_CTX_private *ctx = slot->ctx;
 	PKCS11_TEMPLATE tmpl = {0};
@@ -372,14 +378,14 @@ static int pkcs11_store_key(PKCS11_SLOT_private *slot, EVP_PKEY *pk,
 #else
 	if (pk->type == EVP_PKEY_RSA) {
 		RSA *rsa = pk->pkey.rsa;
-		rsa_n=rsa->n;
-		rsa_e=rsa->e;
-		rsa_d=rsa->d;
-		rsa_p=rsa->p;
-		rsa_q=rsa->q;
-		rsa_dmp1=rsa->dmp1;
-		rsa_dmq1=rsa->dmq1;
-		rsa_iqmp=rsa->iqmp;
+		rsa_n = rsa->n;
+		rsa_e = rsa->e;
+		rsa_d = rsa->d;
+		rsa_p = rsa->p;
+		rsa_q = rsa->q;
+		rsa_dmp1 = rsa->dmp1;
+		rsa_dmq1 = rsa->dmq1;
+		rsa_iqmp = rsa->iqmp;
 #endif
 		pkcs11_addattr_var(&tmpl, CKA_KEY_TYPE, key_type_rsa);
 		pkcs11_addattr_bn(&tmpl, CKA_MODULUS, rsa_n);
@@ -441,6 +447,10 @@ EVP_PKEY *pkcs11_get_key(PKCS11_OBJECT_private *key0, CK_OBJECT_CLASS object_cla
 {
 	PKCS11_OBJECT_private *key = key0;
 	EVP_PKEY *ret = NULL;
+	RSA *rsa;
+#if OPENSSL_VERSION_NUMBER < 0x30000000L || defined(LIBRESSL_VERSION_NUMBER)
+	EC_KEY *ec_key;
+#endif
 
 	if (key->object_class != object_class)
 		key = pkcs11_object_from_object(key, CK_INVALID_HANDLE, object_class);
@@ -451,12 +461,59 @@ EVP_PKEY *pkcs11_get_key(PKCS11_OBJECT_private *key0, CK_OBJECT_CLASS object_cla
 		if (!key->evp_key)
 			goto err;
 	}
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER >= 0x3050000fL )
-	EVP_PKEY_up_ref(key->evp_key);
+	/* We need a full copy of the EVP_PKEY as it will be modified later.
+	 * Using a reference would mean changes to the duplicated EVP_PKEY could
+	 * affect the original one.
+	 */
+	switch (EVP_PKEY_base_id(key->evp_key)) {
+	case EVP_PKEY_RSA:
+		/* Do not try to duplicate foreign RSA keys */
+		rsa = EVP_PKEY_get1_RSA(key->evp_key);
+		if (!rsa)
+			goto err;
+		ret = EVP_PKEY_new();
+		if (!ret) {
+			RSA_free(rsa);
+			goto err;
+		}
+		if (!EVP_PKEY_assign_RSA(ret, rsa)) {
+			RSA_free(rsa);
+			EVP_PKEY_free(ret);
+			goto err;
+		}
+		if (key->object_class == CKO_PRIVATE_KEY)
+			pkcs11_object_ref(key);
+		else /* Public key -> detach PKCS11_OBJECT */
+			pkcs11_set_ex_data_rsa(rsa, NULL);
+		break;
+	case EVP_PKEY_EC:
+#if OPENSSL_VERSION_NUMBER < 0x30000000L || defined(LIBRESSL_VERSION_NUMBER)
+		ec_key = EVP_PKEY_get1_EC_KEY(key->evp_key);
+		if (!ec_key)
+			goto err;
+		ret = EVP_PKEY_new();
+		if (!ret) {
+			EC_KEY_free(ec_key);
+			goto err;
+		}
+		if (!EVP_PKEY_assign_EC_KEY(ret, ec_key)) {
+			EC_KEY_free(ec_key);
+			EVP_PKEY_free(ret);
+			goto err;
+		}
+		if (key->object_class == CKO_PRIVATE_KEY)
+			pkcs11_object_ref(key);
+		else /* Public key -> detach PKCS11_OBJECT */
+			pkcs11_set_ex_data_ec(ec_key, NULL);
 #else
-	CRYPTO_add(&key->evp_key->references, 1, CRYPTO_LOCK_EVP_PKEY);
+		/* pkcs11_ec_copy() method is only set for private keys,
+		 * so public keys do not have a PKCS11_OBJECT reference */
+		ret = EVP_PKEY_dup(key->evp_key);
 #endif
-	ret = key->evp_key;
+		break;
+	default:
+		printf("Unsupported key type\n");
+	}
 err:
 	if (key != key0)
 		pkcs11_object_free(key);
@@ -472,7 +529,7 @@ int pkcs11_authenticate(PKCS11_OBJECT_private *key, CK_SESSION_HANDLE session)
 	PKCS11_SLOT_private *slot = key->slot;
 	PKCS11_CTX_private *ctx = slot->ctx;
 	char pin[MAX_PIN_LENGTH+1];
-	char* prompt;
+	char *prompt;
 	UI *ui;
 	int rv;
 
@@ -517,20 +574,31 @@ int pkcs11_authenticate(PKCS11_OBJECT_private *key, CK_SESSION_HANDLE session)
 }
 
 /*
- * Return keys of a given type (public or private)
+ * Return keys of a given type (public or private) matching the key_template
  * Use the cached values if available
  */
-int pkcs11_enumerate_keys(PKCS11_SLOT_private *slot, unsigned int type,
+int pkcs11_enumerate_keys(PKCS11_SLOT_private *slot, unsigned int type, const PKCS11_KEY *key_template,
 		PKCS11_KEY **keyp, unsigned int *countp)
 {
 	PKCS11_keys *keys = (type == CKO_PRIVATE_KEY) ? &slot->prv : &slot->pub;
+	PKCS11_TEMPLATE tmpl = {0};
 	CK_SESSION_HANDLE session;
+	CK_OBJECT_CLASS object_class = type;
 	int rv;
 
+	pkcs11_addattr_var(&tmpl, CKA_CLASS, object_class);
+	if (key_template) {
+		if (key_template->id_len)
+			pkcs11_addattr(&tmpl, CKA_ID, key_template->id, key_template->id_len);
+
+		if (key_template->label)
+			pkcs11_addattr_s(&tmpl, CKA_LABEL, key_template->label);
+	}
+
 	if (pkcs11_get_session(slot, 0, &session))
 		return -1;
 
-	rv = pkcs11_find_keys(slot, session, type);
+	rv = pkcs11_find_keys(slot, session, type, &tmpl);
 	pkcs11_put_session(slot, session);
 	if (rv < 0) {
 		pkcs11_destroy_keys(slot, type);
@@ -565,21 +633,16 @@ int pkcs11_remove_object(PKCS11_OBJECT_private *obj)
 }
 
 /*
- * Find all keys of a given type (public or private)
+ * Find all keys of a given type (public or private) matching template
  */
-static int pkcs11_find_keys(PKCS11_SLOT_private *slot, CK_SESSION_HANDLE session, unsigned int type)
+static int pkcs11_find_keys(PKCS11_SLOT_private *slot, CK_SESSION_HANDLE session, unsigned int type, PKCS11_TEMPLATE *tmpl)
 {
 	PKCS11_CTX_private *ctx = slot->ctx;
-	CK_OBJECT_CLASS key_search_class;
-	CK_ATTRIBUTE key_search_attrs[1] = {
-		{CKA_CLASS, &key_search_class, sizeof(key_search_class)},
-	};
 	int rv, res = -1;
 
 	/* Tell the PKCS11 lib to enumerate all matching objects */
-	key_search_class = type;
 	rv = CRYPTOKI_call(ctx,
-		C_FindObjectsInit(session, key_search_attrs, 1));
+		C_FindObjectsInit(session, tmpl->attrs, tmpl->nattr));
 	CRYPTOKI_checkerr(CKR_F_PKCS11_FIND_KEYS, rv);
 
 	do {
@@ -611,6 +674,12 @@ static int pkcs11_next_key(PKCS11_CTX_private *ctx, PKCS11_SLOT_private *slot,
 	return 0;
 }
 
+PKCS11_OBJECT_private *pkcs11_object_ref(PKCS11_OBJECT_private *obj)
+{
+	pkcs11_atomic_add(&obj->refcnt, 1, &obj->lock);
+	return obj;
+}
+
 static int pkcs11_init_key(PKCS11_SLOT_private *slot, CK_SESSION_HANDLE session,
 	CK_OBJECT_HANDLE object, CK_OBJECT_CLASS type, PKCS11_KEY **ret)
 {
@@ -665,8 +734,12 @@ void pkcs11_destroy_keys(PKCS11_SLOT_private *slot, unsigned int type)
 
 	while (keys->num > 0) {
 		PKCS11_KEY *key = &keys->keys[--keys->num];
-		if (key->_private)
-			pkcs11_object_free(PRIVKEY(key));
+		PKCS11_OBJECT_private *obj = PRIVKEY(key);
+
+		if (obj) {
+			EVP_PKEY_free(obj->evp_key);
+			pkcs11_object_free(obj);
+		}
 	}
 	if (keys->keys)
 		OPENSSL_free(keys->keys);
