@@ -36,49 +36,6 @@ static int (*orig_pkey_ec_sign) (EVP_PKEY_CTX *ctx,
 	const unsigned char *tbs, size_t tbslen);
 #endif /* OPENSSL_NO_EC */
 
-struct evp_pkey_method_st {
-	int pkey_id;
-	int flags;
-	int (*init) (EVP_PKEY_CTX *ctx);
-	int (*copy) (EVP_PKEY_CTX *dst, EVP_PKEY_CTX *src);
-	void (*cleanup) (EVP_PKEY_CTX *ctx);
-	int (*paramgen_init) (EVP_PKEY_CTX *ctx);
-	int (*paramgen) (EVP_PKEY_CTX *ctx, EVP_PKEY *pkey);
-	int (*keygen_init) (EVP_PKEY_CTX *ctx);
-	int (*keygen) (EVP_PKEY_CTX *ctx, EVP_PKEY *pkey);
-	int (*sign_init) (EVP_PKEY_CTX *ctx);
-	int (*sign) (EVP_PKEY_CTX *ctx, unsigned char *sig, size_t *siglen,
-		const unsigned char *tbs, size_t tbslen);
-	int (*verify_init) (EVP_PKEY_CTX *ctx);
-	int (*verify) (EVP_PKEY_CTX *ctx,
-		const unsigned char *sig, size_t siglen,
-		const unsigned char *tbs, size_t tbslen);
-	int (*verify_recover_init) (EVP_PKEY_CTX *ctx);
-	int (*verify_recover) (EVP_PKEY_CTX *ctx,
-		unsigned char *rout, size_t *routlen,
-		const unsigned char *sig, size_t siglen);
-	int (*signctx_init) (EVP_PKEY_CTX *ctx, EVP_MD_CTX *mctx);
-	int (*signctx) (EVP_PKEY_CTX *ctx, unsigned char *sig, size_t *siglen,
-		EVP_MD_CTX *mctx);
-	int (*verifyctx_init) (EVP_PKEY_CTX *ctx, EVP_MD_CTX *mctx);
-	int (*verifyctx) (EVP_PKEY_CTX *ctx, const unsigned char *sig, int siglen,
-		EVP_MD_CTX *mctx);
-	int (*encrypt_init) (EVP_PKEY_CTX *ctx);
-	int (*encrypt) (EVP_PKEY_CTX *ctx, unsigned char *out, size_t *outlen,
-		const unsigned char *in, size_t inlen);
-	int (*decrypt_init) (EVP_PKEY_CTX *ctx);
-	int (*decrypt) (EVP_PKEY_CTX *ctx, unsigned char *out, size_t *outlen,
-		const unsigned char *in, size_t inlen);
-	int (*derive_init) (EVP_PKEY_CTX *ctx);
-	int (*derive) (EVP_PKEY_CTX *ctx, unsigned char *key, size_t *keylen);
-	int (*ctrl) (EVP_PKEY_CTX *ctx, int type, int p1, void *p2);
-	int (*ctrl_str) (EVP_PKEY_CTX *ctx, const char *type, const char *value);
-} /* EVP_PKEY_METHOD */ ;
-
-#if OPENSSL_VERSION_NUMBER >= 0x10000000L
-#define EVP_PKEY_FLAG_DYNAMIC 1
-#endif
-
 #if OPENSSL_VERSION_NUMBER < 0x10002000L || defined(LIBRESSL_VERSION_NUMBER)
 
 typedef struct {
@@ -197,7 +154,7 @@ static CK_MECHANISM_TYPE pkcs11_md2ckm(const EVP_MD *md)
 		return CKM_SHA512;
 	case NID_sha384:
 		return CKM_SHA384;
-#if !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L && !defined(LIBRESSL_VERSION_NUMBER)
 	case NID_sha3_224:
 		return CKM_SHA3_224;
 	case NID_sha3_256:
@@ -225,7 +182,7 @@ static CK_RSA_PKCS_MGF_TYPE pkcs11_md2ckg(const EVP_MD *md)
 		return CKG_MGF1_SHA512;
 	case NID_sha384:
 		return CKG_MGF1_SHA384;
-#if !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L && !defined(LIBRESSL_VERSION_NUMBER)
 	case NID_sha3_224:
 		return CKG_MGF1_SHA3_224;
 	case NID_sha3_256:
@@ -534,11 +491,6 @@ static EVP_PKEY_METHOD *pkcs11_pkey_method_rsa()
 	new_meth = EVP_PKEY_meth_new(EVP_PKEY_RSA,
 		EVP_PKEY_FLAG_AUTOARGLEN);
 
-#ifdef EVP_PKEY_FLAG_DYNAMIC
-	/* do not allow OpenSSL to free this object */
-	new_meth->flags &= ~EVP_PKEY_FLAG_DYNAMIC;
-#endif
-
 	EVP_PKEY_meth_copy(new_meth, orig_meth);
 
 	EVP_PKEY_meth_set_sign(new_meth,
@@ -676,11 +628,6 @@ static EVP_PKEY_METHOD *pkcs11_pkey_method_ec()
 
 	new_meth = EVP_PKEY_meth_new(EVP_PKEY_EC, 0);
 
-#ifdef EVP_PKEY_FLAG_DYNAMIC
-	/* do not allow OpenSSL to free this object */
-	new_meth->flags &= ~EVP_PKEY_FLAG_DYNAMIC;
-#endif
-
 	EVP_PKEY_meth_copy(new_meth, orig_meth);
 
 	EVP_PKEY_meth_set_sign(new_meth,
@@ -696,11 +643,15 @@ int PKCS11_pkey_meths(ENGINE *e, EVP_PKEY_METHOD **pmeth,
 {
 	static int pkey_nids[] = {
 		EVP_PKEY_RSA,
+#ifndef OPENSSL_NO_EC
 		EVP_PKEY_EC,
+#endif /* OPENSSL_NO_EC */
 		0
 	};
 	static EVP_PKEY_METHOD *pkey_method_rsa = NULL;
+#ifndef OPENSSL_NO_EC
 	static EVP_PKEY_METHOD *pkey_method_ec = NULL;
+#endif /* OPENSSL_NO_EC */
 
 	(void)e; /* squash the unused parameter warning */
 	/* all PKCS#11 engines currently share the same pkey_meths */
